# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
import sys
from magnopy._constants._icons import ICON_IN_ARG


def _get_command_info(args, hide_personal_data=False):
    if hide_personal_data:
        comment = f"Executed the command\n\n  magnopy-optimize-sd {' '.join(sys.argv[1:])}\n\n"
    else:
        comment = f"Executed the command\n\n  {' '.join(sys.argv)}\n\n"

    # Save arguments to the comment
    comment += "which resulted in argument values\n\n" + "\n".join(
        [f"{ICON_IN_ARG} {key:20} : {value}" for key, value in vars(args).items()]
    )

    return comment


def _add_spinham_input(parser):
    parser.add_argument(
        "-ss",
        "--spinham-source",
        type=str,
        # metavar="KEYWORD",
        required=True,
        choices=["GROGU", "TB2J", "tb2j", "grogu"],
        help="Magnopy supports two sources of spin Hamiltonian: GROGU or TB2J. Source of "
        "the spin Hamiltonian can not be deduced reliably from the Hamiltonian file. Thus, "
        "magnopy asks you to specify what was the source of the file that you pass to "
        '"--spinham-filename".',
    )
    parser.add_argument(
        "-sf",
        "--spinham-filename",
        type=str,
        metavar="FILENAME",
        required=True,
        help="Path to the file with the parameters of spin Hamiltonian. If you use "
        '"--spinham-source TB2J", then this is an "exchange.out" file, produced by TB2J. '
        'If you use "--spinham-source GROGU", then this is a .txt file with the '
        "Hamiltonian parameters produced by GROGU.",
    )


def _add_spin_values(parser):
    parser.add_argument(
        "-sv",
        "--spin-values",
        nargs="*",
        type=str,
        metavar=("S1",),
        default=None,
        help='Only works with "--spinham-source TB2J". This option changes spin values of '
        "the magnetic atoms. It expects M numbers in the same order as the order of "
        "magnetic atoms in the input TB2J file. Magnetic atoms are defined as the atoms "
        "that have at least one interaction associated with them. Spin values are always "
        "positive. If you want to specify spin-down state or AFM order use spin directions "
        "and not the spin values.",
    )


def _add_magnetic_field(parser):
    parser.add_argument(
        "-mf",
        "--magnetic-field",
        default=None,
        nargs=3,
        metavar=("B_X", "B_Y", "B_Z"),
        type=float,
        help="Vector of the external magnetic field (magnetic flux density, B), given in "
        "the units of Tesla.",
    )


def _add_output_folder(parser):
    parser.add_argument(
        "-of",
        "--output-folder",
        type=str,
        default="magnopy-results",
        help="Path to the folder where all output files of magnopy wil be saved. Defaults "
        'to "magnopy-results". If the folder does not exist, then magnopy will try to '
        "create it.",
    )


def _add_no_html(parser):
    parser.add_argument(
        "-no-html",
        "--no-html",
        action="store_true",
        default=False,
        help=".html files are generally heavy (~ 5 Mb). This option allows to disable "
        "their generation to save disk space.",
    )


def _add_hide_personal_data(parser):
    parser.add_argument(
        "-hpd",
        "--hide-personal-data",
        action="store_true",
        default=False,
        help="Whether not to use os.path.abspath. Hides the file structure of your "
        "personal computer. Note, that the relative paths can still expose the "
        "structure of you filesystem.",
    )


def _add_supercell(parser):
    parser.add_argument(
        "-s",
        "--supercell",
        nargs=3,
        type=int,
        default=(1, 1, 1),
        metavar=("xA_1", "xA_2", "xA_3"),
        help="Definition of the supercell for the spin optimization. Expects three "
        "integers as an input, each integer is a multiple of the original unit cell "
        "along one of the three lattice vectors. By default optimizes on the unit cell.",
    )


def _add_energy_tolerance(parser):
    parser.add_argument(
        "-et",
        "--energy-tolerance",
        default=1e-5,
        type=float,
        help="Tolerance parameter. Difference between classical energies of two "
        "consecutive optimization steps.",
    )


def _add_torque_tolerance(parser):
    parser.add_argument(
        "-tt",
        "--torque-tolerance",
        default=1e-5,
        type=float,
        help="Tolerance parameter. Maximum value of torque among all spins at current "
        "optimization step.",
    )


def _add_spin_directions(parser):
    parser.add_argument(
        "-sd",
        "--spin-directions",
        nargs="*",
        type=str,
        default=None,
        metavar=("FILENAME",),
        help="Set of spin directions for every magnetic site in the unit cell defines "
        "the ground state of the Hamiltonian. Magnopy can try to identify the ground "
        "state on the fly via energy minimization (see also magnopy-optimize-sd). "
        "However, we recommend to provide the set of spin directions explicitly. This "
        "argument expects a filename for the file that contains three numbers per line "
        "and has M lines in total. M is the amount of magnetic sites. Alternatively, you "
        "can give a set of 3*M numbers directly to this argument. Then, first three "
        "numbers define spin direction of the first magnetic site, next three numbers - "
        "of the second and so on. Magnetic sites are defined as atoms that have at least "
        "one parameter associated with them.",
    )


def _add_k_path(parser):
    parser.add_argument(
        "-kp",
        "--k-path",
        default=None,
        metavar="GAMMA-X-S|GAMMA-Y",
        type=str,
        help="Path in reciprocal space between high symmetry points. Used for the plots "
        'of magnon dispersion and other quantities. Ignored if "--kpoints FILENAME" is '
        "used. List of available high symmetry points is calculated automatically by "
        "wulfric package using convention of HPKOT paper based on the space group of the "
        "crystal structure (see wulfric.org for more details). Numerical tolerance for "
        "the space group search via spglib package can be controlled with "
        '"--spglib-symprec NUMBER" argument. Note: high symmetry points are defined by '
        "wulfric based on the reciprocal lattice of the primitive cell, but their "
        "relative coordinates are given in the basis of reciprocal cell of the input "
        "cell (i. e. the same unit cell as in the TB2J or GROGU file). Primitive cell "
        "might be different from the input cell. If you want to control what atoms are "
        "considered to be equivalent in the space group search - manually change "
        "their names in the input file. To learn how a name is translated into the "
        "spglib_type see documentation of the wulfric package (wulfric.org).",
    )


def _add_k_points(parser):
    parser.add_argument(
        "-kps",
        "--kpoints",
        type=str,
        default=None,
        help="Explicit list of k-points in reciprocal space. A filename is expected. The "
        "file shall contain three numbers per line. Each line specifies one k-point. By "
        "default three numbers are understood as absolute coordinates in reciprocal "
        'space. If "--relative" is used, than three numbers are understood as relative '
        "coordinates in the basis of the reciprocal cell of the input unit cell (i. e. "
        "of the same unit cell as in the TB2J or GROGU file, not of the primitive cell).",
    )


def _add_relative(parser):
    parser.add_argument(
        "-r",
        "--relative",
        default=False,
        action="store_true",
        help='Relevant only if "--kpoints FILENAME" is used. See description of '
        '"--kpoints".',
    )


def _add_number_processors(parser):
    parser.add_argument(
        "-np",
        "--number-processors",
        type=int,
        default=None,
        help="Magnopy is parallelized over the k-points. By default it uses all "
        "available processors. Pass 1 to run in serial.",
    )


def _add_spglib_symprec(parser):
    parser.add_argument(
        "-spg-s",
        "--spglib-symprec",
        metavar="NUMBER",
        type=float,
        default=1e-5,
        help="Tolerance parameter for the space group symmetry search by spglib.",
    )


def _add_spglib_types(parser):
    parser.add_argument(
        "-spg-t",
        "--spglib-types",
        metavar=("TYPE1", "TYPE2"),
        nargs="*",
        type=int,
        default=None,
        help="Set of spglib types for the space group search by spglib. Expects N non-zero "
        "integers in total. N is the amount of all sites/atoms in the input files (not"
        "just magnetic, but all).  If not given, then guesses automatically by wulfric "
        "(see wulfric.get_spglib_types() at wulfric.org types for more details).",
    )


################################################################################
#                                  DEPRECATED                                  #
################################################################################


# DEPRECATED in v0.2.0
# Remove in March 2026
def _add_make_sd_image(parser):
    if sys.version_info >= (3, 13):
        parser.add_argument(
            "-msdi",
            "--make-sd-image",
            nargs=3,
            type=int,
            default=None,
            help="DEPRECATED in v0.2.0. Interactive .html images are generated by default, "
            'use "--no-html" to suppress. This arguments will be removed from magnopy in '
            "March of 2026",
            deprecated=True,
        )
    else:
        parser.add_argument(
            "-msdi",
            "--make-sd-image",
            nargs=3,
            type=int,
            default=None,
            help="DEPRECATED in v0.2.0. Interactive .html images are generated by default, "
            'use "--no-html" to suppress. This arguments will be removed from magnopy in '
            "March of 2026",
        )



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================

import pytest
import numpy as np
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.extra.numpy import arrays as harrays

from magnopy._diagonalization import _check_grand_dynamical_matrix, _inverse_by_colpa


from magnopy import solve_via_colpa
from magnopy import ColpaFailed


@pytest.mark.parametrize(
    "D",
    [
        [[1]],
        [1],
        [1, 2, 3],
        [[[1]]],
        [[1, 2, 3]],
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
    ],
)
def test_check_grand_dynamical_matrix_valid_fails(D):
    with pytest.raises(ValueError):
        _check_grand_dynamical_matrix(D=D)


@pytest.mark.parametrize(
    "D, expected_N",
    [([[1, 0], [0, 1]], 1), (np.ones((6, 6)), 3)],
)
def test_check_grand_dynamical_matrix_valid_passes(D, expected_N):
    D_new, N = _check_grand_dynamical_matrix(D=D)
    assert N == expected_N
    assert np.allclose(D_new, D)


@given(
    D=harrays(
        dtype=np.complex128,
        shape=(2, 2),
        elements=st.complex_numbers(
            min_magnitude=0, max_magnitude=10, allow_subnormal=False
        ),
    )
)
def test_solve_via_colpa_2_x_2(D):
    # Small addition - to avoid dealing with finite precision issues
    D = (D + D.conj().T) / 2 + 1e-10 * np.eye(2)

    try:
        E, G = solve_via_colpa(D)

        E_prime = np.linalg.inv(np.conjugate(G).T) @ D @ np.linalg.inv(G)
        assert np.allclose(np.diag(E), E_prime)

        E_prime = _inverse_by_colpa(np.conjugate(G).T) @ D @ _inverse_by_colpa(G)
        assert np.allclose(np.diag(E), E_prime)

    except ColpaFailed:
        pass



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def read_spin_directions(filename: str):
    r"""
    Reads spin directions from the file.

    Parameters
    ----------

    filename : str
        File with the spin directions. See notes for the details of its content.

    Returns
    -------

    spin_directions : (M, ) :numpy:`ndarray`
        Array with the spin directions. Each row is a spin direction vector normalized to
        1.


    Notes
    -----

    The file is expected to contain three numbers per line, here is an example for two
    spins

    .. code-block:: text

        S1_x S1_y S1_z
        S2_x S2_y S2_z

    Only the direction of the spin vector is recognized, the modulus is ignored. Comments
    are allowed at any place of the file and preceded by the symbol "#". If the symbol "#"
    is found, then the rest of the line is ignored. Here are examples of valid use of the
    comments

    .. code-block:: text

        # Spin vectors for the material XX
        S1_x S1_y S1_z # Atom X1
        # This comments is here by some reason
        S2_x S2_y S2_z # Atom X2

    """

    spin_directions = []
    with open(filename, "r", encoding="utf-8") as f:
        for i, line in enumerate(f):
            # Remove comment lines
            if line.startswith("#"):
                continue
            # Remove inline comments and leading/trailing whitespaces
            line = line.split("#")[0].strip()
            # Check for empty lines
            if line:
                line = line.split()
                if len(line) != 3:
                    raise ValueError(
                        f"Expected three numbers per line (in line {i + 1}), got: {len(line)}."
                    )
                spin_directions.append(list(map(float, line)))

    spin_directions = np.array(spin_directions, dtype=float)
    # Normalize the spin directions
    spin_directions = (
        spin_directions / np.linalg.norm(spin_directions, axis=1)[:, np.newaxis]
    )

    return spin_directions


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================

r"""
Routines for conversion of the full 3x3x3x3 tensor of 422 parameter into its parts and back.
"""

import numpy as np

INDICES = [
    (0, 0, 0, 0),
    (0, 1, 0, 1),
    (0, 2, 0, 2),
    (1, 0, 1, 0),
    (1, 1, 1, 1),
    (1, 2, 1, 2),
    (2, 0, 2, 0),
    (2, 1, 2, 1),
    (2, 2, 2, 2),
]

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def to_biquadratic(parameter, tensor_form=False):
    r"""
    Computes biquadratic exchange parameter from full tensor form.


    .. math::

        C_{4,2,2}
        \sum_{i,j,u,v}
        J^{ijuv}
        S_{\mu}^i
        S_{\mu}^j
        S_{\nu}^u
        S_{\nu}^v
        =
        C_{4,2,2}B
        \left(
        \boldsymbol{S}_{\mu}
        \cdot
        \boldsymbol{S}_{\nu}
        \right)^2
        +
        \dots

    where :math:`B` is defined as

    .. math::

        B = \dfrac{J^{xxxx} + J^{xyxy} + J^{xzxz} +
        J^{yxyx} + J^{yyyy} + J^{yzyz} +
        J^{zxzx} + J^{zyzy} + J^{zzzz}}{9}


    Parameters
    ----------

    parameter : (3, 3, 3, 3) |array-like|_
        Full tensor parameter (:math:`\boldsymbol{J}`).

    tensor_form : bool, default False
        Whether to return tensor form of biquadratic exchange parameter instead of the scalar.

    Returns
    -------

    B : float or (3, 3, 3, 3) :numpy:`ndarray`
        Biquadratic exchange parameter.

        * If ``tensor_form == False``, then returns a number :math:`B`.
        * If ``tensor_form == True``, then returns an array :math:`\boldsymbol{J}_B`.

    See Also
    --------

    from_biquadratic

    Examples
    --------

    .. doctest::

        >>> import numpy as np
        >>> from magnopy import converter422
        >>> parameter = np.ones((3, 3, 3, 3))
        >>> B = converter422.to_biquadratic(parameter=parameter)
        >>> B
        1.0
        >>> J_B = converter422.to_biquadratic(parameter=parameter, tensor_form=True)
        >>> J_B
        array([[[[1., 0., 0.],
                 [0., 0., 0.],
                 [0., 0., 0.]],
        <BLANKLINE>
                [[0., 1., 0.],
                 [0., 0., 0.],
                 [0., 0., 0.]],
        <BLANKLINE>
                [[0., 0., 1.],
                 [0., 0., 0.],
                 [0., 0., 0.]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[0., 0., 0.],
                 [1., 0., 0.],
                 [0., 0., 0.]],
        <BLANKLINE>
                [[0., 0., 0.],
                 [0., 1., 0.],
                 [0., 0., 0.]],
        <BLANKLINE>
                [[0., 0., 0.],
                 [0., 0., 1.],
                 [0., 0., 0.]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[0., 0., 0.],
                 [0., 0., 0.],
                 [1., 0., 0.]],
        <BLANKLINE>
                [[0., 0., 0.],
                 [0., 0., 0.],
                 [0., 1., 0.]],
        <BLANKLINE>
                [[0., 0., 0.],
                 [0., 0., 0.],
                 [0., 0., 1.]]]])

    """

    parameter = np.array(parameter)

    if parameter.shape != (3, 3, 3, 3):
        raise ValueError(f"Wrong shape of the parameter, got {parameter.shape}.")

    B = 0.0

    for i, j, u, v in INDICES:
        B += parameter[i, j, u, v]

    B /= 9

    if tensor_form:
        return from_biquadratic(B=B)

    return float(B)


def from_biquadratic(B):
    r"""
    Computes tensor form of the biquadratic exchange parameter.


    .. math::

        C_{4,2,2}
        B
        \left(
        \boldsymbol{S}_{\mu}
        \cdot
        \boldsymbol{S}_{\nu}
        \right)^2
        =
        C_{4,2,2}
        \sum_{i,j,u,v}
        J_B^{ijuv}
        S_{\mu}^i
        S_{\mu}^j
        S_{\nu}^u
        S_{\nu}^v


    where tensor :math:`\boldsymbol{J}_B` is defined as

    *   :math:`J_B^{ijuv} = B` if :math:`(ijuv)` is one of

        .. math::
            \begin{matrix}
                (xxxx), & (xyxy), & (xzxz), \\
                (yxyx), & (yyyy), & (yzyz), \\
                (zxzx), & (zyzy), & (yyyy)
            \end{matrix}

    *   :math:`J_B^{ijuv} = 0` otherwise.


    Parameters
    ----------

    B : float
        Biquadratic exchange parameter.

    Returns
    -------

    parameter : (3, 3, 3, 3) :numpy:`ndarray`
        Tensor form of the biquadratic exchange parameter.

    See Also
    --------

    to_biquadratic

    Examples
    --------

    .. doctest::

        >>> from magnopy import converter422
        >>> parameter = converter422.from_biquadratic(B=1)
        >>> parameter.shape
        (3, 3, 3, 3)
        >>> parameter
        array([[[[1., 0., 0.],
                 [0., 0., 0.],
                 [0., 0., 0.]],
        <BLANKLINE>
                [[0., 1., 0.],
                 [0., 0., 0.],
                 [0., 0., 0.]],
        <BLANKLINE>
                [[0., 0., 1.],
                 [0., 0., 0.],
                 [0., 0., 0.]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[0., 0., 0.],
                 [1., 0., 0.],
                 [0., 0., 0.]],
        <BLANKLINE>
                [[0., 0., 0.],
                 [0., 1., 0.],
                 [0., 0., 0.]],
        <BLANKLINE>
                [[0., 0., 0.],
                 [0., 0., 1.],
                 [0., 0., 0.]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[0., 0., 0.],
                 [0., 0., 0.],
                 [1., 0., 0.]],
        <BLANKLINE>
                [[0., 0., 0.],
                 [0., 0., 0.],
                 [0., 1., 0.]],
        <BLANKLINE>
                [[0., 0., 0.],
                 [0., 0., 0.],
                 [0., 0., 1.]]]])

    """

    parameter = np.zeros((3, 3, 3, 3), dtype=float)

    for i, j, u, v in INDICES:
        parameter[i, j, u, v] = B

    return parameter


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np
from wulfric.crystal import get_distance

from magnopy._parameters._p22 import from_dmi, from_iso
from magnopy._spinham._convention import Convention
from magnopy._spinham._hamiltonian import SpinHamiltonian

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def load_tb2j(
    filename, spin_values=None, spglib_types=None, g_factors=None, quiet=True
) -> SpinHamiltonian:
    r"""
    Reads spin Hamiltonian from the "exchange.out" file produced by |TB2J|_.

    Parameters
    ----------

    filename : str
        Path to the "exchange.out" file produced by |TB2J|_.

    spin_values : (M, ) iterable of floats, optional
        Spin values for all magnetic atoms. Order is the same as in |TB2J|_ file. Magnetic
        atoms are defined as those that have at least one parameter associated with them.
        If none given, magnopy sets spin value as :math:`|\boldsymbol{m} / g_{factor}|`
        for each spin, where :math:`\boldsymbol{m}` is the magnetic moment, read from the
        |TB2J|_ file.

    spglib_types : (M_prime, ) iterable of ints, optional
        Spglib types for all atoms (not only for magnetic ones, but for all). Order is the
        same as in |TB2J|_ file. If none given, then there will be no "spglib_types" key
        in ``spinham.atoms``.

    g_factors : (M, ) iterable of floats, optional
        g-factors for all atoms. Order is the same as in |TB2J|_ file. If none given, then
        magnopy sets :math:`g = 2` for all atoms.

    quiet : bool, default True
        If ``False``, warnings are printed when distances between atoms computed by
        magnopy are different from the distances read from the |TB2J|_ file. This is a
        legacy feature, kept for the sake of backward compatibility. See Notes for
        details.

    Returns
    -------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian, that is built from the |TB2J|_ file.

    Raises
    ------

    ValueError
        If ``spin_values`` is provided and its length does not match the number of
        magnetic atoms in the Hamiltonian.

    ValueError
        If ``spglib_types`` is provided and its length does not match the number of
        atoms in the Hamiltonian.


    Notes
    -----

    |TB2J|_ outputs distances between atoms for each exchange bond. Magnopy do not use
    those, but rather computes distances based on the unit cell and atom positions if
    necessary. At the moment of reading the file it can check if the distance read from
    the file is different from the computed one and print a warning if ``quiet=False``.
    Computed and read distances tend to differ in the last digits, since the unit cell is
    provided with the same precision as the distance in the |TB2J|_ file.
    """

    minor_sep = "-" * 88
    garbage = str.maketrans(
        {"(": None, ")": None, "[": None, "]": None, ",": None, "'": None}
    )
    # Do not correct spelling, it is taken from TB2J.
    cell_flag = "Cell (Angstrom):"
    atoms_flag = "Atoms:"
    atom_end_flag = "Total"
    exchange_flag = "Exchange:"
    iso_flag = "J_iso:"
    aniso_flag = "J_ani:"
    dmi_flag = "DMI:"

    file = open(filename, "r", encoding="utf-8")
    # model = SpinHamiltonian(convention="TB2J")
    line = True

    # Read everything before exchange
    while line:
        line = file.readline()

        # Read cell
        if line and cell_flag in line:
            a1 = file.readline().split()
            a2 = file.readline().split()
            a3 = file.readline().split()

            cell = np.array(
                [
                    list(map(float, a1)),
                    list(map(float, a2)),
                    list(map(float, a3)),
                ]
            )

        # Read atoms
        if line and atoms_flag in line:
            atoms = dict(names=[], positions=[], magnetic_moments=[], charges=[])
            line = file.readline()
            line = file.readline()
            line = file.readline().split()
            i = 0
            while line and atom_end_flag not in line:
                try:
                    # Slicing is not used intentionally.
                    magmom = tuple(map(float, [line[5], line[6], line[7]]))
                except IndexError:
                    magmom = float(line[5])
                try:
                    charge = float(line[4])
                except IndexError:
                    charge = None

                position = np.array(tuple(map(float, line[1:4]))) @ np.linalg.inv(cell)
                atoms["names"].append(line[0])
                atoms["positions"].append(position)
                atoms["magnetic_moments"].append(magmom)
                atoms["charges"].append(charge)

                line = file.readline().split()
                i += 1

        # Check if the exchange section is reached
        if line and exchange_flag in line:
            break

    # Populate g_factors of atoms
    if g_factors is None:
        g_factors = [2.0 for _ in range(len(atoms["names"]))]

    atoms["g_factors"] = g_factors

    # Create a spin Hamiltonian
    spinham = SpinHamiltonian(
        cell=cell, atoms=atoms, convention=Convention.get_predefined(name="tb2j")
    )

    # Prepare index mapping for atom names
    index_mapping = {}

    # Names of the atoms are unique in the TB2J files
    for index, name in enumerate(spinham.atoms.names):
        index_mapping[name] = index

    # Read exchange (22) parameters
    while line:
        while line and minor_sep not in line:
            line = file.readline()
        line = file.readline().translate(garbage).split()
        atom1 = index_mapping[line[0]]
        atom2 = index_mapping[line[1]]
        ijk = tuple(map(int, line[2:5]))
        distance = float(line[-1])
        iso = None
        aniso = None
        dmi = None
        while line and minor_sep not in line:
            line = file.readline()

            # Read isotropic exchange
            if line and iso_flag in line:
                iso = float(line.split()[-1])

            # Read anisotropic exchange
            if line and aniso_flag in line:
                aniso = np.array(
                    [
                        list(map(float, file.readline().translate(garbage).split())),
                        list(map(float, file.readline().translate(garbage).split())),
                        list(map(float, file.readline().translate(garbage).split())),
                    ]
                )

            # Read DMI
            if line and dmi_flag in line:
                dmi = tuple(map(float, line.translate(garbage).split()[-3:]))

        parameter = np.zeros((3, 3), dtype=float)
        if iso is not None:
            parameter = parameter + from_iso(iso=iso)
        if dmi is not None:
            parameter = parameter + from_dmi(dmi=dmi)
        if aniso is not None:
            parameter = parameter + aniso

        # Adding info from the exchange block to the SpinHamiltonian structure
        spinham.add_22(
            alpha=atom1,
            beta=atom2,
            nu=ijk,
            # Avoid passing aniso to the function as then the function make it traceless
            # and symmetric, potentially loosing part of the matrix.
            # Due to the TB2J problem: aniso not always traceless.
            parameter=parameter,
            when_present="replace",
        )

        computed_distance = get_distance(spinham.cell, spinham.atoms, atom1, atom2, ijk)
        if abs(computed_distance - distance) > 0.001 and not quiet:
            print(
                "\nComputed distance is a different from the read one:\n"
                + f"  Computed: {computed_distance:.4f}\n  "
                + f"Read: {distance:.4f}\n"
            )

    # Populate spin_values of atoms
    if spin_values is not None:
        if len(spin_values) != spinham.M:
            raise ValueError(
                f"Expected {spinham.M} spin values, got {len(spin_values)}"
            )
        true_spin_values = [0.0 for _ in spinham.atoms.names]

        for i in range(len(spin_values)):
            true_spin_values[spinham.map_to_all[i]] = spin_values[i]
    else:
        true_spin_values = [
            float(np.linalg.norm(atoms["magnetic_moments"][alpha]))
            / atoms["g_factors"][alpha]
            for alpha in range(len(atoms["names"]))
        ]

    spinham.atoms["spins"] = true_spin_values

    # Populate spglib_types of atoms
    if spglib_types is not None:
        if len(spglib_types) != len(spinham.atoms.names):
            raise ValueError(
                f"Expected {len(spinham.atoms.names)} spglib types, got {len(spglib_types)}"
            )
        spinham.atoms["spglib_types"] = [int(_) for _ in spglib_types]

    spinham._reset_internals()

    return spinham


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import os

import numpy as np
from wulfric.cell import get_params
from wulfric.crystal import get_atom_species

from magnopy._constants._si import JOULE, ELECTRON_VOLT, MILLI
from magnopy._package_info import logo
from magnopy._parameters._p22 import to_dmi, to_symm_anisotropy
from magnopy._spinham._convention import Convention
from magnopy._spinham._hamiltonian import SpinHamiltonian

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _verified_materials(materials, M):
    if materials is None:
        materials = [i for i in range(M)]
    else:
        if len(materials) != M:
            raise ValueError(f"Expected {M} materials, got {len(materials)}.")
        materials_pool = set(materials)
        higher_material = max(materials_pool)
        for i in range(0, higher_material + 1):
            if i not in materials_pool:
                raise ValueError(
                    f"Materials indices should be consecutive integers between 0 and {higher_material}. Missing {i}."
                )

    return materials


def dump_vampire(
    spinham: SpinHamiltonian,
    seedname="vampire",
    anisotropic=True,
    dmi=True,
    custom_mask=None,
    decimals=5,
    materials=None,
    no_logo=False,
) -> None:
    """
    Saves spin Hamiltonian in the format suitable for |Vampire|_ (.UCF and .mat).

    Parameters
    ----------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian object to be saved.

    seedname : str, default "vampire"
        Seedname for the .UCF and .mat files. Extensions are added automatically. Input
        file is independent of ``seedname`` and always has the name "input-template".

    anisotropic : bool, default True
        Whether to output anisotropic exchange.

    dmi : bool, default True
        Whether to output DMI exchange.

    custom_mask : func, optional
        Custom mask for the exchange parameter. Function that takes (3, 3)
        :numpy:`ndarray` as an input and returns (3, 3) :numpy:`ndarray` as an output. If
        given, then ``anisotropic`` and ``dmi`` parameters are ignored.

        .. code-block:: python

            parameter_to_output = custom_mask(parameter)

    decimals : int, default 4
        Number of decimals to be printed (only for the exchange values).

    materials : list of int, optional
        List of materials for the atoms. Length is the same as the number of magnetic
        atoms in the ``spinham`` (``spinham.M``). Order is the same as in
        :py:attr:`.SpinHamiltonian.magnetic_atoms`. If none given, each magnetic atom is
        considered as a separate material. Material indices start from 0 and should
        contain all consecutive integers between 0 and number of materials. Number of
        materials cannot be higher than number of magnetic atoms.

    no_logo : bool, default False
        Whether to include the logo in the output files.

    Notes
    -----

    Examples of the correct ``materials`` list for 5 magnetic atoms

    .. code-block:: python

        [0, 0, 0, 0, 0]
        [1, 3, 2, 1, 0]
        [0, 1, 2, 3, 4]

    Examples of the incorrect ``materials`` list for 5 magnetic atoms

    .. code-block:: python

        [0, 6, 0, 0, 0]
        [1, 3, 3, 1, 0]
        [1, 2, 3, 4, 5]
    """

    head, _ = os.path.split(seedname)

    if head != "":
        os.makedirs(head, exist_ok=True)

    dump_vampire_ucf(
        spinham,
        filename=f"{seedname}.UCF",
        anisotropic=anisotropic,
        dmi=dmi,
        custom_mask=custom_mask,
        decimals=decimals,
        materials=materials,
        no_logo=no_logo,
    )
    dump_vampire_mat(
        spinham,
        filename=f"{seedname}.mat",
        materials=materials,
        no_logo=no_logo,
    )
    with open(os.path.join(head, "input-template"), "w", encoding="utf-8") as file:
        if not no_logo:
            file.write(f"{logo(comment=True, date_time=True)}\n")

        file.write(
            "\n".join(
                [
                    "#------------------------------------------",
                    f"material:file={seedname}.mat",
                    f"material:unit-cell-file = {seedname}.UCF",
                    "#------------------------------------------",
                    "# TODO: simulation setup",
                ]
            )
        )


def dump_vampire_mat(
    spinham: SpinHamiltonian, filename, materials=None, no_logo=False
) -> None:
    """
    Generates .mat file for |Vampire|_.

    Parameters
    ----------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian object to be saved.

    filename : str
        Name for the .mat file. Extension ".mat" is added if not present.

    materials : list of int, optional
        List of materials for the atoms. Length is the same as the number of magnetic
        atoms in the ``spinham`` (``spinham.M``). Order is the same as in
        :py:attr:`.SpinHamiltonian.magnetic_atoms`. If none given, each magnetic atom is
        considered as a separate material. Material indices start from 0 and should
        contain all consecutive integers between 0 and number of materials. Number of
        materials cannot be higher than number of magnetic atoms.

    no_logo : bool, default False
        Whether to include the logo in the output files.

    Raises
    ------

    ValueError
        If ``materials`` list is given and its length is not equal to the number of
        magnetic atoms in the ``spinham``.

    ValueError
        If ``materials`` list does not contain all consecutive integers between 0 and
        the highest material index.

    Notes
    -----

    Examples of the correct ``materials`` list for 5 magnetic atoms

    .. code-block:: python

        [0, 0, 0, 0, 0]
        [1, 3, 2, 1, 0]
        [0, 1, 2, 3, 4]

    Examples of the incorrect ``materials`` list for 5 magnetic atoms

    .. code-block:: python

        [0, 6, 0, 0, 0]
        [1, 3, 3, 1, 0]
        [1, 2, 3, 4, 5]
    """

    if len(filename) < 4 or filename[-4:] != ".mat":
        filename += ".mat"

    materials = _verified_materials(materials, spinham.M)

    if no_logo:
        text = []
    else:
        text = [logo(comment=True, date_time=True)]

    text.append(f"material:num-materials = {max(materials) + 1}")

    for i, (material, name, spin, g_factor) in enumerate(
        zip(
            materials,
            spinham.magnetic_atoms.names,
            spinham.magnetic_atoms.spins,
            spinham.magnetic_atoms.g_factors,
        )
    ):
        if material not in materials[:i]:
            m_i = material + 1
            text.append("#---------------------------------------------------")
            text.append(f"# Material {m_i}")
            text.append("#---------------------------------------------------")
            text.append(f"material[{m_i}]:material-name = {name}")
            text.append(f"material[{m_i}]:material-element = {get_atom_species(name)}")
            text.append(f"material[{m_i}]:atomic-spin-moment={spin * g_factor} ! muB")
            text.append(f"material[{m_i}]:initial-spin-direction = random")
            text.append(f"material[{m_i}]:damping-constant = 0.1")
            text.append(f"material[{m_i}]:uniaxial-anisotropy-constant = 0.0")

    text.append("#---------------------------------------------------")

    with open(filename, "w", encoding="utf-8") as file:
        file.write("\n".join(text))


def dump_vampire_ucf(
    spinham: SpinHamiltonian,
    filename,
    anisotropic=True,
    dmi=True,
    custom_mask=None,
    decimals=5,
    materials=None,
    no_logo=False,
) -> None:
    """
    Generates .UCF file for |Vampire|_.

    Parameters
    ----------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian object to be saved.

    filename : str, optional
        Name for the .UCF file. Extension ".UCF" is added if not present.

    anisotropic : bool, default True
        Whether to output anisotropic exchange.

    dmi : bool, default True
        Whether to output DMI exchange.

    custom_mask : func, optional
        Custom mask for the exchange parameter. Function that takes (3, 3)
        :numpy:`ndarray` as an input and returns (3, 3) :numpy:`ndarray` as an output. If
        given, then ``anisotropic`` and ``dmi`` parameters are ignored.

        .. code-block:: python

            parameter_to_output = custom_mask(parameter)

    decimals : int, default 4
        Number of decimals to be printed (only for the exchange values).

    materials : list of int, optional
        List of materials for the atoms. Length is the same as the number of magnetic
        atoms in the ``spinham`` (``spinham.M``). Order is the same as in
        :py:attr:`.SpinHamiltonian.magnetic_atoms`. If none given, each magnetic atom is
        considered as a separate material. Material indices start from 0 and should
        contain all consecutive integers between 0 and number of materials. Number of
        materials cannot be higher than number of magnetic atoms.

    no_logo : bool, default False
        Whether to include the logo in the output files.

    Raises
    ------

    ValueError
        If ``materials`` list is given and its length is not equal to the number of
        magnetic atoms in the ``spinham``.

    ValueError
        If ``materials`` list does not contain all consecutive integers between 0 and
        the highest material index.

    Notes
    -----

    Examples of the correct ``materials`` list for 5 magnetic atoms

    .. code-block:: python

        [0, 0, 0, 0, 0]
        [1, 3, 2, 1, 0]
        [0, 1, 2, 3, 4]

    Examples of the incorrect ``materials`` list for 5 magnetic atoms

    .. code-block:: python

        [0, 6, 0, 0, 0]
        [1, 3, 3, 1, 0]
        [1, 2, 3, 4, 5]
    """

    if len(filename) < 4 or filename[-4:] != ".UCF":
        filename += ".UCF"

    materials = _verified_materials(materials, spinham.M)

    original_convention = spinham.convention
    spinham.convention = Convention.get_predefined(name="Vampire")

    if no_logo:
        text = []
    else:
        text = [logo(comment=True, date_time=True)]

    a, b, c, _, _, _ = get_params(spinham.cell)
    text.append("# Unit cell size:")
    text.append(f"{a:.8f} {b:.8f} {c:.8f}")
    text.append("# Unit cell lattice vectors:")
    text.append(
        f"{spinham.cell[0][0]:15.8f} {spinham.cell[0][1]:15.8f} {spinham.cell[0][2]:15.8f}"
    )
    text.append(
        f"{spinham.cell[1][0]:15.8f} {spinham.cell[1][1]:15.8f} {spinham.cell[1][2]:15.8f}"
    )
    text.append(
        f"{spinham.cell[2][0]:15.8f} {spinham.cell[2][1]:15.8f} {spinham.cell[2][2]:15.8f}"
    )
    text.append("# Atoms")
    text.append(f"{len(spinham.magnetic_atoms.names)} {len(np.unique(materials))}")

    for alpha in range(spinham.M):
        position = spinham.magnetic_atoms.positions[alpha]
        text.append(
            f"{alpha:<5} {position[0]:15.8f} {position[1]:15.8f} {position[2]:15.8f} {materials[alpha]:>5}"
        )

    text.append("# Interactions")
    text.append(f"{len(spinham.p22)} tensorial")

    IID = 0
    fmt = f"{7 + decimals}.{decimals}e"

    # Write (two spins & one site)
    for alpha, J in spinham.p21:
        alpha = spinham.map_to_magnetic[alpha]
        if custom_mask is not None:
            J = custom_mask(J)
        else:
            if not dmi:
                J -= to_dmi(J, matrix_form=True)
            if not anisotropic:
                J -= to_symm_anisotropy(J)
        J = J * (MILLI * ELECTRON_VOLT) / JOULE
        text.append(
            f"{IID:<5} {alpha:>3} {alpha:>3}  {0:>2} {0:>2} {0:>2}  "
            f"{J[0][0]:{fmt}} {J[0][1]:{fmt}} {J[0][2]:{fmt}} "
            f"{J[1][0]:{fmt}} {J[1][1]:{fmt}} {J[1][2]:{fmt}} "
            f"{J[2][0]:{fmt}} {J[2][1]:{fmt}} {J[2][2]:{fmt}}"
        )
        IID += 1

    # Write (two spins & two sites)
    bonds = []
    for alpha, beta, nu, J in spinham.p22:
        alpha = spinham.map_to_magnetic[alpha]
        beta = spinham.map_to_magnetic[beta]
        if custom_mask is not None:
            J = custom_mask(J)
        else:
            if not dmi:
                J -= to_dmi(J, matrix_form=True)
            if not anisotropic:
                J -= to_symm_anisotropy(J)
        # print(alpha, beta, nu)
        # print(J, end="\n\n")
        J = J * (MILLI * ELECTRON_VOLT) / JOULE
        r_alpha = np.array(spinham.magnetic_atoms.positions[alpha])
        r_beta = np.array(spinham.magnetic_atoms.positions[beta])

        distance = np.linalg.norm((r_beta - r_alpha + nu) @ spinham.cell)
        bonds.append([alpha, beta, nu, J, distance])

    bonds = sorted(bonds, key=lambda x: x[4])
    for alpha, beta, (i, j, k), J, _ in bonds:
        text.append(
            f"{IID:<5} {alpha:>3} {beta:>3}  {i:>2} {j:>2} {k:>2}  "
            f"{J[0][0]:{fmt}} {J[0][1]:{fmt}} {J[0][2]:{fmt}} "
            f"{J[1][0]:{fmt}} {J[1][1]:{fmt}} {J[1][2]:{fmt}} "
            f"{J[2][0]:{fmt}} {J[2][1]:{fmt}} {J[2][2]:{fmt}}"
        )
        IID += 1

    spinham.convention = original_convention

    with open(filename, "w", encoding="utf-8") as file:
        file.write("\n".join(text))


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np
import pytest
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.extra.numpy import arrays as harrays

from magnopy import span_local_rf


def test_span_local_rf_along_z():
    direction = [0, 0, 1]

    x_a, y_a, z_a = span_local_rf(direction)

    assert np.allclose(x_a, [1, 0, 0])
    assert np.allclose(y_a, [0, 1, 0])
    assert np.allclose(z_a, [0, 0, 1])


def test_span_local_rf_opposite_z():
    direction_vector = [0, 0, -1]

    x_a, y_a, z_a = span_local_rf(direction_vector)

    assert np.allclose(x_a, [0, -1, 0])
    assert np.allclose(y_a, [-1, 0, 0])
    assert np.allclose(z_a, [0, 0, -1])


@given(
    harrays(
        np.float64,
        (3,),
        elements=st.floats(min_value=-1e8, max_value=1e8, allow_subnormal=False),
    )
)
def test_span_local_rf(direction_vector):
    if np.allclose(direction_vector, np.zeros(3)):
        with pytest.raises(ValueError):
            x_a, y_a, z_a = span_local_rf(direction_vector)

    else:
        x_a, y_a, z_a = span_local_rf(direction_vector)

        assert np.linalg.det([x_a, y_a, z_a]) > 0.0

        assert np.allclose(
            [np.linalg.norm(x_a), np.linalg.norm(y_a), np.linalg.norm(z_a)], np.ones(3)
        )

        assert np.allclose([x_a @ y_a, x_a @ z_a, y_a @ z_a], np.zeros(3))



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


R"""
Convention of spin Hamiltonian
"""

from magnopy._spinham._hamiltonian import SpinHamiltonian

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def make_supercell(spinham: SpinHamiltonian, supercell):
    r"""
    Creates a spin Hamiltonian on the supercell.

    Parameters
    ----------

    spinham : :py:class:`.SpinHamiltonian`
        Original spin Hamiltonian. ``spinham.cell`` is interpreted as the original
        unit cell.

    supercell : (3, ) tuple or list of int
        Repetitions of the unit cell (``spinham.cell``) along each lattice
        vector that define the unit cell. If :math:`(i, j, k)` is given, then the
        supercell is formally defined as
        :math:`(i\cdot\boldsymbol{a}_1, j\cdot\boldsymbol{a}_2, k\cdot\boldsymbol{a}_3)`,
        where :math:`(\boldsymbol{a}_1, \boldsymbol{a}_2, \boldsymbol{a}_3)` is the
        original cell (``spinham.cell``).

    Returns
    -------

    new_spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian that is defined on a ``supercell`` and has the same parameters
        as the given ``spinham`` propagated over the whole supercell.

    Examples
    --------

    First a simple example with only on-site parameters present in the Hamiltonian (i.e
    ``p21``) and two atoms per unit cell.

    .. doctest::

        >>> import numpy as np
        >>> import magnopy
        >>> # First create the original spin Hamiltonian
        >>> cell = np.eye(3)
        >>> atoms = dict(
        ...     names=["Fe1", "Fe2"],
        ...     positions=[[0, 0, 0], [0.5, 0.5, 0.5]],
        ...     spins=[1, 1],
        ...     g_factors=[2, 2],
        ... )
        >>> convention = magnopy.Convention(
        ...     spin_normalized=False, multiple_counting=True, c21=1
        ... )
        >>> spinham = magnopy.SpinHamiltonian(
        ...     cell=cell, atoms=atoms, convention=convention
        ... )
        >>> # Add an on-site parameter for both atoms
        >>> spinham.add_21(alpha=0, parameter=np.eye(3))
        >>> spinham.add_21(alpha=1, parameter=np.eye(3))
        >>> # Now create a spin Hamiltonian on the (2, 2, 2) supercell
        >>> new_spinham = magnopy.make_supercell(spinham=spinham, supercell=(2, 2, 2))
        >>> # Note that the unit cell of the new Hamiltonian is the supercell of the original one
        >>> new_spinham.cell
        array([[2., 0., 0.],
               [0., 2., 0.],
               [0., 0., 2.]])
        >>> # All the atoms of the supercell are now contained in the unit cell of the
        >>> # new Hamiltonian
        >>> len(spinham.atoms.names)
        2
        >>> len(new_spinham.atoms.names)
        16
        >>> for i in range(2):
        ...     print(
        ...         spinham.atoms.names[i],
        ...         spinham.atoms.positions[i],
        ...         spinham.atoms.spins[i],
        ...     )
        Fe1 [0, 0, 0] 1
        Fe2 [0.5, 0.5, 0.5] 1
        >>> for i in range(16):
        ...     print(
        ...         new_spinham.atoms.names[i],
        ...         new_spinham.atoms.positions[i],
        ...         new_spinham.atoms.spins[i],
        ...     )
        Fe1_0_0_0 [0.0, 0.0, 0.0] 1
        Fe2_0_0_0 [0.25, 0.25, 0.25] 1
        Fe1_1_0_0 [0.5, 0.0, 0.0] 1
        Fe2_1_0_0 [0.75, 0.25, 0.25] 1
        Fe1_0_1_0 [0.0, 0.5, 0.0] 1
        Fe2_0_1_0 [0.25, 0.75, 0.25] 1
        Fe1_1_1_0 [0.5, 0.5, 0.0] 1
        Fe2_1_1_0 [0.75, 0.75, 0.25] 1
        Fe1_0_0_1 [0.0, 0.0, 0.5] 1
        Fe2_0_0_1 [0.25, 0.25, 0.75] 1
        Fe1_1_0_1 [0.5, 0.0, 0.5] 1
        Fe2_1_0_1 [0.75, 0.25, 0.75] 1
        Fe1_0_1_1 [0.0, 0.5, 0.5] 1
        Fe2_0_1_1 [0.25, 0.75, 0.75] 1
        Fe1_1_1_1 [0.5, 0.5, 0.5] 1
        Fe2_1_1_1 [0.75, 0.75, 0.75] 1
        >>> # The parameters are updated as well
        >>> len(spinham.p21)
        2
        >>> for alpha, _ in spinham.p21:
        ...     print(alpha)
        0
        1
        >>> len(new_spinham.p21)
        16
        >>> for alpha, _ in new_spinham.p21:
        ...     print(alpha)
        0
        1
        2
        3
        4
        5
        6
        7
        8
        9
        10
        11
        12
        13
        14
        15

    Second example with one atom per unit cell, but with the bilinear interaction that
    connects two different atoms (``p22``).

    .. doctest::

        >>> import numpy as np
        >>> import magnopy
        >>> # First create the original spin Hamiltonian
        >>> cell = np.eye(3)
        >>> atoms = dict(names=["Fe"], positions=[[0, 0, 0]], spins=[1], g_factors=[2])
        >>> convention = magnopy.Convention(
        ...     spin_normalized=False, multiple_counting=True, c22=1
        ... )
        >>> spinham = magnopy.SpinHamiltonian(
        ...     cell=cell, atoms=atoms, convention=convention
        ... )
        >>> # Add an on-site parameter for both atoms
        >>> spinham.add_22(alpha=0, beta=0, nu=(1, 0, 0), parameter=np.eye(3))
        >>> # Now create a spin Hamiltonian on the (2, 2, 2) supercell
        >>> new_spinham = magnopy.make_supercell(spinham=spinham, supercell=(2, 2, 2))
        >>> len(new_spinham.atoms.names)
        8
        >>> for i in range(8):
        ...     print(
        ...         new_spinham.atoms.names[i],
        ...         new_spinham.atoms.positions[i],
        ...         new_spinham.atoms.spins[i],
        ...     )
        Fe_0_0_0 [0.0, 0.0, 0.0] 1
        Fe_1_0_0 [0.5, 0.0, 0.0] 1
        Fe_0_1_0 [0.0, 0.5, 0.0] 1
        Fe_1_1_0 [0.5, 0.5, 0.0] 1
        Fe_0_0_1 [0.0, 0.0, 0.5] 1
        Fe_1_0_1 [0.5, 0.0, 0.5] 1
        Fe_0_1_1 [0.0, 0.5, 0.5] 1
        Fe_1_1_1 [0.5, 0.5, 0.5] 1
        >>> # The bonds were recalculated automatically
        >>> for alpha, beta, nu, _ in spinham.p22:
        ...     print(alpha, beta, nu)
        0 0 (1, 0, 0)
        0 0 (-1, 0, 0)
        >>> for alpha, beta, nu, _ in new_spinham.p22:
        ...     print(alpha, beta, nu)
        0 1 (0, 0, 0)
        1 0 (1, 0, 0)
        2 3 (0, 0, 0)
        3 2 (1, 0, 0)
        4 5 (0, 0, 0)
        5 4 (1, 0, 0)
        6 7 (0, 0, 0)
        7 6 (1, 0, 0)
        1 0 (0, 0, 0)
        0 1 (-1, 0, 0)
        3 2 (0, 0, 0)
        2 3 (-1, 0, 0)
        5 4 (0, 0, 0)
        4 5 (-1, 0, 0)
        7 6 (0, 0, 0)
        6 7 (-1, 0, 0)


    """

    if supercell[0] < 1 or supercell[1] < 1 or supercell[2] < 1:
        raise ValueError(
            f"Supercell repetitions should be larger or equal to 1, got {supercell}"
        )

    new_cell = [supercell[i] * spinham.cell[i] for i in range(3)]

    new_atoms = {}

    for key in spinham.atoms:
        new_atoms[key] = []

    for k in range(supercell[2]):
        for j in range(supercell[1]):
            for i in range(supercell[0]):
                for atom_index in range(len(spinham.atoms.names)):
                    for key in spinham.atoms:
                        if key == "positions":
                            position = spinham.atoms.positions[atom_index]
                            new_position = [
                                (position[0] + i) / supercell[0],
                                (position[1] + j) / supercell[1],
                                (position[2] + k) / supercell[2],
                            ]
                            new_atoms["positions"].append(new_position)
                        elif key == "names":
                            new_atoms["names"].append(
                                f"{spinham.atoms.names[atom_index]}_{i}_{j}_{k}"
                            )
                        else:
                            new_atoms[key].append(spinham.atoms[key][atom_index])

    new_spinham = SpinHamiltonian(
        cell=new_cell, atoms=new_atoms, convention=spinham.convention
    )

    def get_new_indices(alpha, nu, ijk):
        nu = [nu[index] + ijk[index] for index in range(3)]

        i, j, k = [nu[index] % supercell[index] for index in range(3)]

        nu = [nu[index] // supercell[index] for index in range(3)]

        alpha = alpha + (i + j * supercell[0] + k * supercell[1] * supercell[0]) * len(
            spinham.atoms.names
        )

        return alpha, tuple(nu)

    for k in range(supercell[2]):
        for j in range(supercell[1]):
            for i in range(supercell[0]):
                # One spin
                for alpha, parameter in spinham._1:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))

                    new_spinham.add_1(alpha=alpha, parameter=parameter)

                # Two spins
                for alpha, parameter in spinham._21:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))

                    new_spinham.add_21(alpha=alpha, parameter=parameter)

                for alpha, beta, nu, parameter in spinham._22:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))
                    beta, nu = get_new_indices(alpha=beta, nu=nu, ijk=(i, j, k))

                    new_spinham.add_22(
                        alpha=alpha,
                        beta=beta,
                        nu=nu,
                        parameter=parameter,
                        when_present="replace",
                    )

                # Three spins
                for alpha, parameter in spinham._31:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))

                    new_spinham.add_31(alpha=alpha, parameter=parameter)

                for alpha, beta, nu, parameter in spinham._32:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))
                    beta, nu = get_new_indices(alpha=beta, nu=nu, ijk=(i, j, k))

                    new_spinham.add_32(
                        alpha=alpha,
                        beta=beta,
                        nu=nu,
                        parameter=parameter,
                        when_present="replace",
                    )

                for alpha, beta, gamma, nu, _lambda, parameter in spinham._33:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))
                    beta, nu = get_new_indices(alpha=beta, nu=nu, ijk=(i, j, k))
                    gamma, _lambda = get_new_indices(
                        alpha=gamma, nu=_lambda, ijk=(i, j, k)
                    )

                    new_spinham.add_33(
                        alpha=alpha,
                        beta=beta,
                        gamma=gamma,
                        nu=nu,
                        _lambda=_lambda,
                        parameter=parameter,
                        when_present="replace",
                    )

                # Four spins
                for alpha, parameter in spinham._41:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))

                    new_spinham.add_41(alpha=alpha, parameter=parameter)

                for alpha, beta, nu, parameter in spinham._421:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))
                    beta, nu = get_new_indices(alpha=beta, nu=nu, ijk=(i, j, k))

                    new_spinham.add_421(
                        alpha=alpha,
                        beta=beta,
                        nu=nu,
                        parameter=parameter,
                        when_present="replace",
                    )

                for alpha, beta, nu, parameter in spinham._422:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))
                    beta, nu = get_new_indices(alpha=beta, nu=nu, ijk=(i, j, k))

                    new_spinham.add_422(
                        alpha=alpha,
                        beta=beta,
                        nu=nu,
                        parameter=parameter,
                        when_present="replace",
                    )

                for alpha, beta, gamma, nu, _lambda, parameter in spinham._43:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))
                    beta, nu = get_new_indices(alpha=beta, nu=nu, ijk=(i, j, k))
                    gamma, _lambda = get_new_indices(
                        alpha=gamma, nu=_lambda, ijk=(i, j, k)
                    )

                    new_spinham.add_43(
                        alpha=alpha,
                        beta=beta,
                        gamma=gamma,
                        nu=nu,
                        _lambda=_lambda,
                        parameter=parameter,
                        when_present="replace",
                    )

                for (
                    alpha,
                    beta,
                    gamma,
                    epsilon,
                    nu,
                    _lambda,
                    rho,
                    parameter,
                ) in spinham._44:
                    alpha, _ = get_new_indices(alpha=alpha, nu=(0, 0, 0), ijk=(i, j, k))
                    beta, nu = get_new_indices(alpha=beta, nu=nu, ijk=(i, j, k))
                    gamma, _lambda = get_new_indices(
                        alpha=gamma, nu=_lambda, ijk=(i, j, k)
                    )
                    epsilon, rho = get_new_indices(alpha=epsilon, nu=rho, ijk=(i, j, k))
                    new_spinham.add_44(
                        alpha=alpha,
                        beta=beta,
                        gamma=gamma,
                        epsilon=epsilon,
                        nu=nu,
                        _lambda=_lambda,
                        rho=rho,
                        parameter=parameter,
                        when_present="replace",
                    )

    return new_spinham


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir
